from OritatamiSimNodes import *
from SimulatorFramework import SimUtils
import math
import os
from collections import deque
import copy
import itertools
import random


class ONodeGrid:
    def __init__(self, width, height):
        self.MakeGrid(width, height)
        self.delayLength = 2    # Default delay of 2
        self.arity = 1          # Maximum number of bonds for a single bead
        self.lockWithoutBinding = False
        self.bindFullFloppySegment = False
        self.maxFloppyLength = 9    # Maximum length of a floppy segment to test during simulation before terminating
        self.bAllDone = False
        self.seed = []
        self.beadCycle = []
        self.origBeadCycle = []
        self.currCyclePos = 0
        self.lastBeadLoc = None
        self.lastSeedLoc = None
        self.floppySegment = deque([])
        self.ONodeTypes = dict()
        self._numBeads = 0

    def GetNumBeads(self):
        return self._numBeads
    
    def GetNumTiles(self):
        return self._numBeads

    def GetWidth(self):
        return self.width
    
    def GetHeight(self):
        return self.height

    def MakeGrid(self, width, height, bResetBeadNum=False):
        self.width = width
        self.height = height
        self._grid = [[None for x in range(width)] for y in range(height)]
        self.minX = 0 - int(math.floor(width/2))
        self.maxX = int(math.floor(width/2))
        self.minY = 0 - int(math.floor(height/2))
        self.maxY = int(math.floor(height/2))
        if (width % 2) == 0:
            self.minX = self.minX + 1
        if (height % 2) == 0:
            self.minY = self.minY + 1
        if bResetBeadNum == True:
            self._numBeads = 0

    def GetLoc(self, coords):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            print("Growing grid because of access to " + str(coords))
            self.GrowGrid()
        return self._grid[coords[0]-self.minX][coords[1]-self.minY]
    
    def SetLoc(self, coords, onode):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            self.GrowGrid()
        self._grid[coords[0]-self.minX][coords[1]-self.minY] = onode
        return True

    def AddBond(self, coords1, coords2):
        loc1 = self.GetLoc(coords1)
        loc2 = self.GetLoc(coords2)
        if loc1 is None or loc2 is None:
            if loc1 is None:
                print("ERROR: trying to create a bond between " + str(coords1) + " and " + str(coords2) + " but " + str(coords1) + " is empty")
            else:
                print("ERROR: trying to create a bond between " + str(coords1) + " and " + str(coords2) + " but " + str(coords2) + " is empty")
            return False
        if loc1.AddBond(coords2) > -1:
            if loc2.AddBond(coords1) > -1:
                return True
        return False
    
    def RemoveBond(self, coords1, coords2):
        if self.GetLoc(coords1).RemoveBond(coords2) > -1:
            if self.GetLoc(coords2).RemoveBond(coords1) > -1:
                return True
        return False

    def GrowGrid(self):
        oldgrid = self._grid
        oldwidth = self.width
        oldheight = self.height
        oldminx = self.minX
        oldminy = self.minY
        self.MakeGrid(oldwidth + GRIDGROWTHINC, oldheight + GRIDGROWTHINC)
        for i in range(oldwidth):
            for j in range(oldheight):
                oldnode = oldgrid[i][j]
                self.SetLoc((i + oldminx, j + oldminy), oldnode)

    def AddNode(self, otype, coords, parentCoords, bWarnOnNoParent=True):
        childNode = ONode(otype, coords)
        if not self.GetLoc(coords) is None:
            print("ERROR: Trying to add another node at location " + str(coords))
            return
        self.SetLoc(coords, childNode)
        self.SetParent(coords, parentCoords, bWarnOnNoParent)
        self.lastBeadLoc = coords
        self._numBeads = self._numBeads + 1
        
    def SetParent(self, coords, parentCoords, bWarnOnNoParent=True):
        childNode = self.GetLoc(coords)
        if self.GetLoc(coords) is None:
            print("ERROR: Trying to set parent of non-existent node at location " + str(coords))
            return
        parentNode = None
        if (not parentCoords is None):
            parentNode = self.GetLoc(parentCoords)
        elif bWarnOnNoParent == True:
            print("Adding a node with no parent at location " + str(coords))
        if (not parentNode is None) and (not parentNode.child is None):
            print("Warning: Parent node already has a child, so not setting parent")
        else:
            childNode.parent = parentNode
            if not parentNode is None:
                parentNode.child = childNode
        
    def RemoveNode(self, coords):
        node = self.GetLoc(coords)
        if not node is None:
            if node.parent is not None:
                node.parent.child = None
                if self.lastBeadLoc == coords:
                    self.lastBeadLoc = node.parent.coords
            elif self.lastBeadLoc == coords:
                self.lastBeadLoc = None
            if node.child is not None:
                node.child.parent = None
            bonds = node._bonds #MJP: abusing private member here...
            for b in bonds:
                self.RemoveBond(coords, b)
            self.SetLoc(coords, None)
            self._numBeads = self._numBeads - 1
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))


    def AreNbrLocs(self, coords1, coords2):
        if (coords1 is None) or (coords2 is None):
            return False
        for n in ONBRS:
            nbrCoords = (coords1[0]+n[0], coords1[1]+n[1])
            if nbrCoords == coords2:
                return True
        return False

    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in ONBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist

    def GetOccupiedNbrLocs(self, coords, bSkipParent=False, maxBonds=0):
        retlist = []
        for n in ONBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                if (bSkipParent == False) or (nbrNode.child is None) or (not nbrNode.child.coords == coords):
                    if nbrNode.NumBonds() <= maxBonds:
                        retlist.append(nbrNode)
        return retlist


    # Note: make sure beadsToAdd is reversed, i.e. the first bead to add is the last in the list!!!
    def GetPathsFreeToDistance(self, start, beadsToAdd):
        #print("GetPathsFreeToDistance called with start " + str(start) + ", beadsToAdd " + str(beadsToAdd))
        maxConns = 0
        maxConnsTo = []
        maxConnsFrom = []
        bestPaths = []

        nbrCoordSet = self.GetNbrLocs(start.coords, True)
        if len(nbrCoordSet) == 0:
            return (0, [[]], [[]], [[]])

        for nbrCoords in nbrCoordSet:
            
            poppedBead = beadsToAdd.popleft()
            newType = self.ONodeTypes[poppedBead]

            # We've found a free location to place the next bead
            nbrNode = ONode(newType, nbrCoords, start)
            
            # Temporarily add the bead at this location
            self.AddNode(poppedBead, nbrCoords, start.coords)

            # Going forward without binding it to anything
            if len(beadsToAdd) > 0:
                (conns, connsTo, connsFrom, paths) = self.GetPathsFreeToDistance(nbrNode, beadsToAdd)
                
                if conns >= maxConns:
                    for p in paths:
                        p.insert(0, nbrCoords)
                        #p.append(nbrCoords)

                if conns == maxConns:
                    maxConnsTo = maxConnsTo + connsTo
                    maxConnsFrom = maxConnsFrom + connsFrom
                    bestPaths = bestPaths + paths
                if conns > maxConns:
                    maxConns = conns
                    maxConnsTo = connsTo
                    maxConnsFrom = connsFrom
                    bestPaths = paths

            # For each bead that it can bind to, create that bond and go forward
            maxInnerConns=0
            maxInnerConnsTo=[]
            maxInnerConnsFrom=[]
            bestInnerPaths=[]

            nbrBound = False
            occupiedNbrs = self.GetOccupiedNbrLocs(nbrCoords, True, self.arity-1)
            #########################
            bondableNbrs = []
            for on in occupiedNbrs:
                if (newType.BondsToNodeType(on.type) == True):
                    bondableNbrs.append(on.coords)

            for i in range(1,self.arity+1):
                nbrSets = list(itertools.combinations(bondableNbrs, i))
                
                for nbrSet in nbrSets:
                    #print("for nbrCoords " + str(nbrCoords) + "nbrSet " + str(nbrSet))

                # See if nodes can bind and already existing node hadn't already bound to something (or is linked)
                #if (newType.BondsToNodeType(on.type) == True):
                    nbrBound = True
                    conns = 0
                    connsTo = [[]]
                    connsFrom = [[]]
                    paths = [[]]
                    
                    #self.AddBond(nbrCoords, on.coords)
                    for nbr in nbrSet:
                        self.AddBond(nbrCoords, nbr)

                    # Now go forward with the bond
                    if len(beadsToAdd) > 0:
                        (conns, connsTo, connsFrom, paths) = self.GetPathsFreeToDistance(nbrNode, beadsToAdd)
                    
                    #self.RemoveBond(nbrCoords, on.coords)
                    for nbr in nbrSet:
                        self.RemoveBond(nbrCoords, nbr)
                        for ct in connsTo:
                            ct.insert(0, nbrCoords)
                        for cf in connsFrom:
                            cf.insert(0, nbr)
                    conns = conns + len(nbrSet)
                    #===========================================================
                    # for ct in connsTo:
                    #     ct.insert(0, nbrCoords)
                    # for cf in connsFrom:
                    #     cf.insert(0, on.coords)
                    #===========================================================
                        
                    if conns >= maxInnerConns:
                        for p in paths:
                            p.insert(0, nbrCoords)
                            #p.append(nbrCoords)
                        
                    # Do a check for local max here
                    if conns == maxInnerConns:
                        maxInnerConnsTo = maxInnerConnsTo + connsTo
                        maxInnerConnsFrom = maxInnerConnsFrom + connsFrom
                        bestInnerPaths = bestInnerPaths + paths
                    if conns > maxInnerConns:
                        maxInnerConns = conns
                        maxInnerConnsTo = connsTo
                        maxInnerConnsFrom = connsFrom
                        bestInnerPaths = paths                    
            
            if nbrBound == False and len(beadsToAdd) == 0:  # Case where this is the last bead but couldn't bind to anything
                if maxConns == 0:
                    #bestPaths = list(bestPaths)
                    bestPaths.append([nbrCoords])
                    maxConnsTo.append([])
                    maxConnsFrom.append([])
            else:
                #print(" bestPaths after inner: " + str(bestPaths))
                if maxInnerConns == maxConns:
                    maxConnsTo = maxConnsTo + maxInnerConnsTo
                    maxConnsFrom = maxConnsFrom + maxInnerConnsFrom
                    bestPaths = bestPaths + bestInnerPaths
                if maxInnerConns > maxConns:
                    maxConns = maxInnerConns
                    maxConnsTo = maxInnerConnsTo
                    maxConnsFrom = maxInnerConnsFrom
                    bestPaths = bestInnerPaths
                #print(" bestPaths after combo: " + str(bestPaths))

            # Remove the bead that we temporarily added
            self.RemoveNode(nbrCoords)
            beadsToAdd.extendleft([poppedBead])
            
            #print(" bestPaths after loop iteration: " + str(bestPaths))

        #print("GetPathsFreeToDistance returning with bestPaths " + str(bestPaths))
        return (maxConns, maxConnsTo, maxConnsFrom, bestPaths)
        
    def InitializeFloppySegment(self):
        self.floppySegment = deque([])
        self.currentCyclePos = 0
        if len(self.beadCycle) == 0:
            SimUtils.globalApp.OutputMessage("WARNING: Bead cycle is empty!")
            return False
        for i in range(self.delayLength-1):
            self.floppySegment.append(self.beadCycle[i % len(self.beadCycle)])
            #self.currCyclePos = ((self.currCyclePos + 1) % len(self.beadCycle))
        print("In InitializeFloppySegment, floppy segment = " + str(self.floppySegment))

    def StepForward(self):
        retStat = ORI_STAT_OK
        retList = []

        if len(self.beadCycle) == 0:
            SimUtils.globalApp.ShowWarningDialog("WARNING: Bead cycle is empty!  Cannot simulate an assembly step.")
            SimUtils.globalApp.OutputMessage("WARNING: Bead cycle is empty!  Cannot simulate an assembly step.")
            return (ORI_STAT_NO_SPACE, [])
        
        self.floppySegment.append(self.beadCycle[(self.currCyclePos+len(self.floppySegment)) % len(self.beadCycle)])
        if self.bindFullFloppySegment == True:
            while len(self.floppySegment) < self.delayLength:
                self.floppySegment.append(self.beadCycle[(self.currCyclePos+len(self.floppySegment)) % len(self.beadCycle)])

        if self.maxFloppyLength < len(self.floppySegment):
            SimUtils.globalApp.ShowWarningDialog("The floppy segment has exceeded the maximum\nfloppy segment length setting (currently set to " + str(self.maxFloppyLength) + ").\nStopping simulation.")
            SimUtils.globalApp.OutputMessage("The floppy segment has exceeded the maximum floppy segment length setting (currently set to " + str(self.maxFloppyLength) + ").  Stopping simulation.")
            return (ORI_STAT_MAX_FLOPPY, [])

        SimUtils.PrintDebug("Testing addition of beads: " + str(self.floppySegment) + " starting from location " + str(self.lastBeadLoc), SimUtils.DEBUG_MSG)

        if self.lastBeadLoc is None:
            SimUtils.globalApp.ShowWarningDialog("There is no starting bead location.  Stopping simulation.")
            SimUtils.globalApp.OutputMessage("There is no starting bead location.  Stopping simulation.")
            return (ORI_STAT_NO_SPACE, [])
        startNode = self.GetLoc(self.lastBeadLoc)
        nextBeads = deque(self.floppySegment)
        #nextBeads.reverse()
        SimUtils.PrintDebug("About to call GetPathsFreeToDistance with startNode " + str(startNode) + " and nextBeads " + str(nextBeads), SimUtils.DEBUG_MSG)
        if SimUtils.globalApp.bShowDebugMsgs == True:
            SimUtils.globalApp.OutputMessage("From " + str(startNode.coords) + " checking possible configurations for: " + str(nextBeads))
        (maxNumBonds, maxConnsTo, maxConnsFrom, bestPaths) = self.GetPathsFreeToDistance(startNode, nextBeads)
        if (len(bestPaths) == 0) or (len(bestPaths[0]) == 0):
            SimUtils.globalApp.ShowWarningDialog("There are no free locations left to place beads.  Stopping simulation.")
            SimUtils.globalApp.OutputMessage("There are no free locations left to place beads.  Stopping simulation.")
            return (ORI_STAT_NO_SPACE, [])
        #cProfile.runctx('self.GetPathsFreeToDistance(startNode, 10, [])', None, locals())
        SimUtils.PrintDebug("Printing results from GetPathsFreeToDistance:", SimUtils.DEBUG_MSG)
        SimUtils.PrintDebug("maxConnsTo: " + str(maxConnsTo) + ", maxConnsFrom: " + str(maxConnsFrom) + ", bestPaths: " + str(bestPaths), SimUtils.DEBUG_MSG)
        SimUtils.PrintDebug("maxNumBonds: " + str(maxNumBonds), SimUtils.DEBUG_MSG)
        #for i in range(len(bestPaths)):
        #    print("best path: " + str(bestPaths[i]) + ", connsTo: " + str(maxConnsTo[i]) + ", connsFrom: " + str(maxConnsFrom[i]))

        if self.lockWithoutBinding == False and maxNumBonds == 0:
            SimUtils.PrintDebug("NO BINDING!!!!", SimUtils.DEBUG_MSG)
            listShow = list(nextBeads)
            SimUtils.globalApp.OutputMessage("From " + str(startNode.coords) + " no configurations with bonds found for " + str(listShow))
            return (ORI_STAT_NO_BINDING, [])
        else:
            pathsToAdd = []
            duplicateIndices = []
            count = 0
            for bp in bestPaths:
                path = []
                if self.bindFullFloppySegment == True:
                    numToAdd = len(self.floppySegment)
                    if len(bp) < numToAdd:
                        numToAdd = len(bp)
                    for i in range(numToAdd):
                        path.append(bp[i])
                else:
                    numToAdd = (len(self.floppySegment) - self.delayLength) + 1
                    if len(bp) < numToAdd:
                        numToAdd = len(bp)
                    for i in range(numToAdd):
                        path.append(bp[i])
                if not path in pathsToAdd:
                    pathsToAdd.append(path)
                else:
                    duplicateIndices.append(count)
                count = count + 1
            #print("Duplicate indices: " + str(duplicateIndices))
                
            loc = 0
            if len(pathsToAdd) > 1:
                SimUtils.globalApp.OutputMessage("Nondeterminism:  there are " + str(len(pathsToAdd)) + " paths with equal bonds: " )
                outStr = ""
                for outBP in pathsToAdd:
                    outStr = outStr + str(outBP)
                SimUtils.globalApp.OutputMessage(outStr)
                retStat = ORI_STAT_NONDET
                loc = random.randint(0, len(pathsToAdd)-1)
                SimUtils.globalApp.OutputMessage("Nondeterministically selected path " + str(loc) + ": " + str(pathsToAdd[loc]))
    
            parentLoc = self.lastBeadLoc
            pathToAdd = pathsToAdd[loc]
            
            # Check to see if this path is able to place all beads that it's supposed to
            bTooShort = False
            if self.bindFullFloppySegment == True:
                numToAdd = len(self.floppySegment)
                if len(pathToAdd) < numToAdd:
                    bTooShort = True
            else:
                numToAdd = (len(self.floppySegment) - self.delayLength) + 1
                if len(pathToAdd) < numToAdd:
                    bTooShort = True
            if bTooShort:
                SimUtils.globalApp.OutputMessage("There are not enough free locations left to place all beads.")
            
            for b in pathToAdd:
                newNodeType = self.floppySegment.popleft()
                retList.append(b)
                self.AddNode(newNodeType, b, parentLoc)
                self.currCyclePos = ((self.currCyclePos + 1) % len(self.beadCycle))
                parentLoc = b
                SimUtils.PrintDebug("Adding node of type " + newNodeType + " at " + str(b), SimUtils.DEBUG_MSG)
                SimUtils.PrintDebug("Floppy segment now " + str(self.floppySegment), SimUtils.DEBUG_MSG)
                
            adjLoc = loc
            for d in duplicateIndices:
                if d <= adjLoc:
                    adjLoc = adjLoc + 1
            for b in pathToAdd:
                # Check to see if the new node bonds to existing beads and add bond if so
                #MJP: currently going to lock in secondary structure
                for i in range(len(maxConnsTo[adjLoc])):
                    (x2,y2) = maxConnsTo[adjLoc][i]
                    (x3,y3) = maxConnsFrom[adjLoc][i]
                    #if ((x,y) == (x2,y2)) and ((x3,y3) not in bestPaths[adjLoc]):
                    if (b == (x2,y2)):
                        if self.AddBond(b, (x3,y3)) == False:
                            SimUtils.globalApp.ShowErrorDialog("ERROR: Failed to add bond between beads at " + str(b) + " and " + str((x3,y3)) +" \nStopping simulation.")
                            SimUtils.globalApp.OutputMessage("ERROR: Failed to add bond between beads at " + str(b) + " and " + str((x3,y3)) +" \nStopping simulation.")
                            return (ORI_STAT_ERROR, [])
        return (retStat, retList)

    def LoadOritatamiSystem(self, origFileName, origFileNameFullPath=None):
        fileName = origFileName
        if not origFileNameFullPath is None:
            fileName = origFileNameFullPath
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        if not os.path.isfile(fileName):
            SimUtils.globalApp.OutputMessage("ERROR: Oritatami system definition file '" + fileName + "' not found or not accessible.")
            SimUtils.globalApp.ShowErrorDialog("ERROR: Oritatami system definition file '" + fileName + "' not found or not accessible.")
            return

        rFile = open(fileName, "r")
        print("Using oritatami system definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()

        self.ONodeTypes = dict()

        for line in lines:
            if line.startswith("BeadTypes="):
                try:
                    beadTypes = eval(line[len("BeadTypes="):])
                    for bead in beadTypes:
                        beadType = ONodeType(bead[0], bead[1], bead[2])
                        self.ONodeTypes[bead[0]] = beadType
                    self.FixBeadBondSymmetry(True)
                except:
                    SimUtils.globalApp.ShowErrorDialog("Invalid bead type definitions.  See messages window.")
                    SimUtils.globalApp.OutputMessage("Invalid line '" + line + "'.  Bead types should be a comma separated list surrounded by '[' and ']'. Each bead type definition should similar to the following:  ['Name', 'Color', ['Name2', 'Name3']] where Name2 and Name3 are in a list of bead types that this type binds to.")
                    return False
            elif line.startswith("DelayLength="):
                dlLine = line[len("DelayLength="):].strip()
                self.delayLength = int(dlLine)
                print("DelayLength: " + str(self.delayLength))
                #if (self.delayLength > 2):
                    #MJP TODO: fix this issue!
                    #print("\n\nWARNING:  For delay longer than 2, newly extended beads could possibly bind to each other, but this is not yet implemented!!!\n")
            elif line.startswith("Arity="):
                arityLine = line[len("Arity="):].strip()
                self.arity = int(arityLine)
                print("Arity: " + str(self.arity))
            elif line.startswith("LockWithoutBinding="):
                lwbLine = line[len("LockWithoutBinding="):].strip()
                self.lockWithoutBinding = False
                if lwbLine == "True":
                    self.lockWithoutBinding = True
                print("LockWithoutBinding: " + str(self.lockWithoutBinding))
            elif line.startswith("BindFullFloppySegment="):
                bffsLine = line[len("BindFullFloppySegment="):].strip()
                self.bindFullFloppySegment = False
                if bffsLine == "True":
                    self.bindFullFloppySegment = True
                print("BindFullFloppySegment:" + str(self.bindFullFloppySegment))
            elif line.startswith("BeadCycle="):
                bcLine = line[len("BeadCycle="):].strip()
                self.beadCycle = []
                self.origBeadCycle = []
                try:
                    self.beadCycle = eval(bcLine)
                    self.origBeadCycle = self.beadCycle
                except:
                    SimUtils.globalApp.ShowErrorDialog("Invalid bead cycle definition.  See messages window.")
                    SimUtils.globalApp.OutputMessage("Invalid line '" + line + "'.  Bead cycle should be a comma separated list of bead type names surrounded by '[' and ']', similar to the following:  ['Name1', 'Name2', 'Name3'].")
                    return False
                print("beadCycle: " + str(self.beadCycle))

        SimUtils.globalApp.menuSettings.Check(SimUtils.globalApp.itemSetLWB.GetId(), self.lockWithoutBinding)
        SimUtils.globalApp.menuSettings.Check(SimUtils.globalApp.itemSetBFFS.GetId(), self.bindFullFloppySegment)
        self.InitializeFloppySegment()
        SimUtils.globalApp.ResetBreakpoints()
        SimUtils.globalApp.systemDefinitionFile = origFileName
        SimUtils.globalApp.systemDefinitionFileFullPath = origFileNameFullPath
        SimUtils.globalApp.OutputMessage("Loaded bead type definitions from file '" + fileName + "'")
        print("Loaded bead set: " + str(self.ONodeTypes))
        return True;

        #except:
        #    SimUtils.globalApp.ShowErrorDialog("\nERROR: Invalid oritatami system definition file: '" + fileName + "'\n")
        #    return False
        
    # Check for asymmetry in definitions of which bead types each can bind to
    def FixBeadBondSymmetry(self, bShowWarnings=True, setName=None, setList=None):
        retVal = False
        if not setName is None:
            btNames = self.ONodeTypes.keys()
            others = [item for item in btNames if item not in setList]
            for other in others:
                otherBT = self.ONodeTypes[other]
                if setName in otherBT.bondToTypes:
                    otherBT.bondToTypes.remove(setName)
                    retVal = True
        mismatches = []
        for btName in self.ONodeTypes:
            bt1 = self.ONodeTypes[btName]
            for bondable1 in bt1.bondToTypes:
                bt2 = self.ONodeTypes[bondable1]
                if btName not in bt2.bondToTypes:
                    mismatches.append((bt2.name, bt1.name))
        if len(mismatches) > 0:
            if bShowWarnings == True:
                SimUtils.globalApp.ShowWarningDialog("Invalid bead type definitions automatically repaired.  See messages window.")
            outStr = "The following pairs of bead types were only defined to bind in one of the bead type definitions.  They have been modified so both types include each other.\n"
            for mm in mismatches:
                outStr = outStr + mm[0] + " to " + mm[1] + "\n"
                self.ONodeTypes[mm[0]].bondToTypes.append(mm[1])
            if bShowWarnings == True:
                SimUtils.globalApp.OutputMessage(outStr)
            return True
        else:
            return retVal

    def LoadOritatamiAssembly(self, origFileName, bForceDefLoad=False):
        fileName = origFileName
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        lastLoc = None

        if not os.path.isfile(fileName):
            SimUtils.globalApp.OutputMessage("ERROR: Oritatami assembly definition file '" + fileName + "' not found or not accessible.")
            SimUtils.globalApp.ShowErrorDialog("ERROR: Oritatami assembly definition file '" + fileName + "' not found or not accessible.")
            return
            
        rFile = open(fileName, "r")
        print("Using oritatami assembly definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()
        
        for line in lines:
            if line.startswith("OritatamiSystemDef="):
                defFile = line[len("OritatamiSystemDef="):].strip()
                if (bForceDefLoad == True) or (SimUtils.globalApp.systemDefinitionFile is None) or (not defFile == SimUtils.globalApp.systemDefinitionFile):
                    defFileFullPath = defFile
                    if os.path.isabs(defFileFullPath) == False:
                        dirPath = os.path.dirname(fileName)
                        defFileFullPath = os.path.join(dirPath, defFileFullPath)
                    if not self.LoadOritatamiSystem(defFile, defFileFullPath) == True:
                        return False
            elif line.startswith("Assembly="):
                if self.LoadOritatamiAssemblyHelper(line) == False:
                    SimUtils.globalApp.OutputMessage("ERROR: Invalid oritatami assembly definition file: '" + fileName + "'")
                    return False
            elif line.startswith("LastBead="):
                lastLocStr = line[len("LastBead="):].strip()
                lastLoc = eval(lastLocStr)
        #except:
        #    SimUtils.globalApp.ShowErrorDialog("\nERROR: Invalid oritatami assembly definition file: '" + fileName + "'")
        #    return False
        
        if not lastLoc is None:
            if self.GetLoc(lastLoc) is None:
                SimUtils.globalApp.OutputMessage("ERROR: Assembly definition has invalid last bead location " + str(lastLoc))
            else:
                self.lastSeedLoc = lastLoc
                self.lastBeadLoc = lastLoc
                
        SimUtils.globalApp.seedDefinitionFile = origFileName
        SimUtils.globalApp.OutputMessage("Loaded assembly definition from file '" + SimUtils.globalApp.seedDefinitionFile + "'")
        return True

    def LoadOritatamiAssemblyHelper(self, seedString):
        seedString = seedString[len("Assembly="):].strip()
        try:
            self.seed = []
            seedArr = eval(seedString)
            for b in seedArr:
                if b[0] in self.ONodeTypes:
                    self.seed.append(b)
                else:
                    SimUtils.globalApp.OutputMessage("ERROR: Trying to add a bead of undefined type '" + str(b[0]) + "' at " + str(b[1]))
        except:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: Invalid oritatami assembly definition file.")
            return False
        return self.ResetToSeed()
        
    def ResetToSeed(self):
        print("Seed: " + str(self.seed))
        self.MakeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE, True)
        savedLastSeedLoc = self.lastSeedLoc
        for b in self.seed:
            if b[0] in self.ONodeTypes:
                self.AddNode(b[0], b[1], None, False)
                self.lastSeedLoc = b[1]
            else:
                SimUtils.globalApp.OutputMessage("ERROR: Trying to add a bead of undefined type '" + str(b[0]) + "' at " + str(b[1]))
        if not savedLastSeedLoc is None:
            self.lastSeedLoc = savedLastSeedLoc
        # If a parent is not specified, the previous bead is used by default.  To specify no parent, use "None"
        parent = None
        for b in self.seed:
            if len(b) > 2:
                parent = b[2]
            self.SetParent(b[1], parent)
            parent = b[1]

        for b in self.seed:
            if len(b) > 3:
                blist = b[3]
                for bond in blist:
                    #MJP: check to see if these types should actually bond and warn if not!
                    t1 = self.GetLoc(b[1])
                    t2 = self.GetLoc(bond)
                    if (not t1 is None) and (not t2 is None):
                        if t1.type in self.ONodeTypes:
                            t1type = self.ONodeTypes[t1.type]
                            if t1type.BondsToNodeType(t2.type) == False:
                                SimUtils.globalApp.OutputMessage("WARNING: Creating a bond between non-binding bead types '" + t1.type + "' at " + str(b[1]) + " and '" + t2.type + "' at " + str(bond))
                        self.AddBond(b[1], bond)
        self.InitializeFloppySegment()
        self.lastBeadLoc = self.lastSeedLoc
        self.currCyclePos = 0
        self.bAllDone = False
        SimUtils.globalApp.ResetBreakpoints()
        SimUtils.globalApp.canvas.SendStatusBarUpdate(2, "Number of beads: " + str(self.GetNumBeads()))
        return True

