import random
import os
import ast
import TileType
import Frontier
import SimUtils


GRID_WIDTH  = 100
GRID_HEIGHT = 100
GROWTH_INC  = 200

N  = 0
NE = 1
E  = 2
SE = 3
S  = 4
SW = 5
W  = 6
NW = 7
ALLDIRS = 8
DIRSTRS = ("N", "NE", "E", "SE", "S", "SW", "W", "NW")
NBRS = [(0,1),(1,1),(1,0),(1,-1),(0,-1),(-1,-1),(-1,0),(-1,1)]



class Assembly:
    def __init__(self, width=GRID_WIDTH, height=GRID_HEIGHT):
        #globalApp.OutputMessage("Creating assembly with width=" + str(width) + " and height=" + str(height))
        self.width = width
        self.height = height
        self.minX = -int(width/2)
        self.maxX = int(width/2)
        self.minY = -int(height/2)
        self.maxY = int(height/2)
        self.grid = [[None for x in range(width+1)] for y in range(height+1)]
        self.tileTypes = dict()        
        self.seedPatch = []
        self.frontier = set()
        self.numTiles = 0
        self.systemDefFile = None
        self.tileSystemDef = None
        #globalApp.OutputMessage("Created new grid with (minX,minY) = " + str((self.minX,self.minY)) + " and (maxX,maxY) = " + str((self.maxX,self.maxY)))

    def GetWidth(self):
        return self.width

    def GetHeight(self):
        return self.height

    def GetNumTiles(self):
        return self.numTiles

    def GrowSpace(self):
        newWidth = self.width + GROWTH_INC
        newHeight = self.height + GROWTH_INC
        newAssembly = Assembly(newWidth, newHeight)

        for i in range(self.minX, self.maxX+1):
            for j in range(self.minY, self.maxY+1):
                loc = self.GetLoc((i,j))
                if loc is not None:
                    newAssembly.SetLoc((i,j), loc.type)
        del self.grid[:]
        self.grid = newAssembly.grid
        self.width = newAssembly.width
        self.height = newAssembly.height
        self.minX = -int(newAssembly.width/2)
        self.maxX = int(newAssembly.width/2)
        self.minY = -int(newAssembly.height/2)
        self.maxY = int(newAssembly.height/2)
        
    def ClearSpace(self):
        newWidth = GRID_WIDTH
        newHeight = GRID_HEIGHT
        newAssembly = Assembly(newWidth, newHeight)
        del self.grid[:]
        self.grid = newAssembly.grid
        self.width = newAssembly.width
        self.height = newAssembly.height
        self.minX = -int(newAssembly.width/2)
        self.maxX = int(newAssembly.width/2)
        self.minY = -int(newAssembly.height/2)
        self.maxY = int(newAssembly.height/2)
        self.frontier = set()
        self.numTiles = 0
        
    def GetLoc(self, coords, growIfNeeded=False):
        if (coords[0] < self.minX) or (coords[0] > self.maxX):
            if growIfNeeded == True:
                self.GrowSpace()
            else:
                return None
        if (coords[1] < self.minY) or (coords[1] > self.maxY):
            if growIfNeeded == True:
                self.GrowSpace()
            else:
                return None
        return self.grid[coords[0]-self.minX][coords[1]-self.minY]
    
    def SetLoc(self, coords, ttype):
        if (coords[0] < self.minX) or (coords[0] > self.maxX):
            SimUtils.PrintDebug("ERROR: Attempted to grow beyond grid boundaries!", SimUtils.ERROR_MSG)
            return False
        if (coords[1] < self.minY) or (coords[1] > self.maxY):
            SimUtils.PrintDebug("ERROR: Attempted to grow beyond grid boundaries!", SimUtils.ERROR_MSG)
            return False
        self.grid[coords[0]-self.minX][coords[1]-self.minY] = TileType.Tile(ttype, coords)
        return True
        
    def ClearLoc(self, coords):
        if (coords[0] < self.minX) or (coords[0] > self.maxX):
            return False
        if (coords[1] < self.minY) or (coords[1] > self.maxY):
            return False
        self.grid[coords[0]-self.minX][coords[1]-self.minY] = None
        return True
        
    def AddTile(self, ttype, coords):
        loc = self.GetLoc(coords)
        if not loc is None:
            SimUtils.PrintDebug("ERROR: Trying to add another node at location " + str(coords), SimUtils.ERROR_MSG)
            return False
        if self.SetLoc(coords, ttype) == False:
            return False
        emptyNbrLocs = self.GetNbrLocs(coords, True)
        #print("unoccupied nbrs of " + str(coords) + ": " + str(emptyNbrLocs))
        for n in emptyNbrLocs:   # loop over unoccupied nbr locations to find occupied nbrs of each and build a list of fitting types
            self.AddToFrontierIfNecessary(n)
        self.numTiles = self.numTiles + 1
        return True

    def AddToFrontierIfNecessary(self, coords):
        fittingNbrSet = self.GetFittingTilesForLoc(coords)
        if fittingNbrSet is None:
            return
        else:
            fLoc = Frontier.FrontierLocation(coords, fittingNbrSet)
            self.frontier.discard(fLoc)
            self.frontier.add(fLoc)

    def RemoveTile(self, coords):
        if not self.GetLoc(coords) is None:
            self.ClearLoc(coords)
            self.numTiles = self.numTiles - 1

            # Check to see if this location should now be added to the frontier
            self.AddToFrontierIfNecessary(coords)

            # Check to see if nbrs were in the frontier but should now be removed
            nbrLocs = self.GetNbrLocs(coords, True)
            for n in nbrLocs:
                fLoc = Frontier.FrontierLocation(n, None)
                self.frontier.discard(fLoc)
                self.AddToFrontierIfNecessary(n)
            return True
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))
            return False


    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist


    def GetOccupiedNbrLocs(self, coords):
        retlist = []
        for n in NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                retlist.append(nbrNode)
        return retlist
    

    # This is just a stubbed out function to show how to gather up info about neighbors, and must be completed
    # to work correctly for the system being simulated
    def GetFittingTilesForLoc(self, coords):
        retSet = None
        count = 0
        for n in NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrTile = self.GetLoc(nbrCoords, True)
            if not nbrTile is None:
                oppDir = (count + 4) % 8
                #TODO: Fill in code based on the system being simulated to determine what tile types can go here!!!
            count = count + 1
        return retSet

    def StepForward(self):
        # Pick a spot from the frontier, add a tile
        if len(self.frontier) == 0:
            SimUtils.globalApp.OutputMessage("The frontier is empty!")
            return None
        else:
            frontierLoc = random.sample(self.frontier, 1)[0]
            self.frontier.remove(frontierLoc)
            typeName = frontierLoc.GetRandomFittingTileTypeName()
            if self.AddTile(self.tileTypes[typeName], frontierLoc.coords) == True:
                SimUtils.PrintDebug("Added a tile of type " + typeName + " to location " + str(frontierLoc.coords), SimUtils.MSG_MSG)
                return (frontierLoc.coords, typeName)
            SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            return None
                
    def PlaceSeed(self):
        SimUtils.globalApp.OutputMessage("Placing seed patch: " + str(self.seedPatch))
        for seedTile in self.seedPatch:
            floc = Frontier.FrontierLocation(seedTile[1],"")
            if floc in self.frontier:
                self.frontier.remove(floc)
            self.AddTile(self.tileTypes[seedTile[0]], seedTile[1])
        
    # This is just a stubbed out version of a function to load an assembly system from file(s).  It must be completed
    # for your particular systems and model.
    def LoadTileSystem(self, fileName):
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        #try:
        os.stat(fileName)
        rFile = open(fileName, "r")
        SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()
        
        self.systemDefFile = fileName
        self.tileSystemDef = ast.literal_eval(lines[0])
        print("Read tile system " + str(self.tileSystemDef))

        self.tileTypes = dict()
        self.seedPatch = []
        self.frontier = set()
               
        SimUtils.globalApp.OutputMessage("Loaded tile set with size: " + str(len(self.tileTypes)))


        self.PlaceSeed()
        
        return True

